""" Resource Manager provides a base class for manipulating all of the resources in the database
It contains an abstract _create method that has a base implementation for deserializing the common elements of the resources. This method should be extended by 

:Author: Bilal Shaikh < bilalshaikh42@gmail.com >
:Date: 2020-01-06
:Copyright: 2020, Karr Lab
:License: MIT
"""

from abc import ABC, abstractmethod
from bson.objectid import ObjectId
import pymongo
from biosimulations_query.utils import IdExistsError, DoesNotExistError, MongoUtil
import datetime
from .utils import genAccessToken
from .utils import parseTime
from biosimulations_query.utils.errors import UserDoesNotExistError


class ResourceManager(ABC):
    def __init__(
            self,
            username,
            password,
            host,
            authSource,
            database,
            collection,
            user_manager):
        self.mongo_manager = MongoUtil(
            host=host,
            username=username,
            password=password,
            authSource=authSource,
            db=database,
        )
        self.client, self.db, self.collection = self.mongo_manager.con_db(
            collection)
        self.user_manager = user_manager

    def _parse_body(self, body):
        """Takes in a dictionary representing a resource and manipulates it to the correct format for the database. 
        Usernames and resource ID's are converted to ObjectIDs for the mongo database

        Args:
            body (dict): A dictionary of attributes ofor the resource

        Raises:
            UserDoesNotExistError: The username provided does not exists in the user database
            IdExistsError: The chosen id for a resource already exists in the database
            DoesNotExistError: The id provided for a resource does not exists in the database
        Returns:
            dict: The resource after processing
        """
        userName = body.get("owner")
        if(userName):
            owner = self.user_manager.get_id_from_username(userName)
            if (owner):
                body["owner"] = owner
            else:
                raise UserDoesNotExistError(userName)
        return body

    def _parse_response(self, response):
        """ Gets a dictionary from the Mongo database and converts the ObjectIDs to user set IDs.         
        Args:
            response (dict): The response from the mongo query

        Returns:
            dict: The resource in external format
        """
        userId = response.get("owner")
        if(userId):
            owner = self.user_manager.get_username_from_id(userId)
            response["owner"] = owner
        return response

    def _parse_responses(self, responses):
        """Takes a list of resources from the database and calls _parse_response on each member

        Args:
            responses (list): A list of resource objects

        Returns:
            list: A list of resource objects in external format
        """
        new_responses = []
        for response in responses:
            new_responses.append(self._parse_response(response))
        return new_responses

    def _get_user_id(self, username):
        """Gets the Auth0 Id of a user given the username. The auth0 id is also the ObjectID in the database

        Args:
            username (string): The username of the user

        Returns:
            string: The Auth0 id of the user, which is also the primary key
        """
        return self.user_manager.get_id_from_username(username)

    @abstractmethod
    def _create(self, **kwargs):
        resourceId = kwargs.get("id", None)
        parent = kwargs.get('parent', None)
        name = kwargs.get("name", None)
        image = kwargs.get("image", None)
        owner = kwargs.get("owner", "bilalshaikh42")
        description = kwargs.get("description")
        summary = kwargs.get("summary")
        private = kwargs.get("private", False)
        published = kwargs.get("published", False)
        tags = kwargs.get("tags", [])
        created = kwargs.get("created")
        if (created):
            created = parseTime(created)
        else:
            created = datetime.datetime.utcnow()
        updated = datetime.datetime.utcnow()
        identifiers = kwargs.get("identifiers", [])
        accessToken = kwargs.get("accessToken", genAccessToken())
        license = kwargs.get("license", "mit")
        references = kwargs.get('references', [])
        authors = kwargs.get(
            "authors", [{"firstName": owner, "lastName": "", "middleName": ""}])
        resource = {
            "id": resourceId,
            "name": name,
            "image": image,
            "owner": owner,
            "description": description,
            "private": private,
            "published": published,
            "summary": summary,
            "tags": tags,
            "created": created,
            "updated": updated,
            "identifiers": identifiers,
            "accessToken": accessToken,
            "license": license,
            "authors": authors,
            "parent": parent,
            "references": references}
        resource = self._parse_body(resource)
        return resource

    def add_one(self, **kwargs):
        resource = self._create(**kwargs)
        _id = ObjectId()

        if resource["id"] is None:
            resource["id"] = str(_id)
            resource["_id"] = _id

        try:
            resource = self.collection.insert_one(resource)
        except pymongo.errors.DuplicateKeyError:
            raise IdExistsError()

        resource = self.collection.find_one(
            {"_id": resource.inserted_id}, {'_id': False})
        resource = self._parse_response(resource)
        return resource

    def get_all(self):
        resources = list(self.collection.find({}, {'_id': False}))
        resources = self._parse_responses(resources)
        return resources

    def get_all_public(self):
        resources = list(self.collection.find(
            {"private": False}, {'_id': False}))
        resources = self._parse_responses(resources)
        return resources

    def get_private_by_owner(self, owner):
        owner = self._get_user_id(owner)
        resources = list(self.collection.find(
            {"private": True, "owner": owner}, {'_id': False}))
        resources = self._parse_responses(resources)
        return resources

    def get_by_owner(self, owner):
        owner = self._get_user_id(owner)
        resources = list(self.collection.find(
            {"owner": owner}, {'_id': False}))
        resources = self._parse_responses(resources)
        return resources

    def get_public_by_owner(self, owner):
        owner = self._get_user_id(owner)
        resources = list(self.collection.find(
            {"private": False, "owner": owner}, {'_id': False}))
        resources = self._parse_responses(resources)
        return resources

    def get_by_identifier(self, identifier):
        resource = self.collection.find_one({"id": identifier}, {'_id': False})
        if (resource is None):
            raise DoesNotExistError(identifier)
        resource = self._parse_response(resource)
        return resource

    def update_one(self, identifier, **kwargs):
        resource = self.collection.find_one({"id": identifier})

        if resource is None:
            raise DoesNotExistError(identifier)

        # Don't override created
        kwargs["created"] = resource.get(
            "created")

        _id = resource["_id"]
        updatedResource = self._create(**kwargs)

        try:
            self.collection.replace_one({"_id": _id}, updatedResource)
        except pymongo.errors.DuplicateKeyError:
            raise IdExistsError(identifier)

        resource = self.collection.find_one({"_id": _id}, {'_id': False})

        if resource is None:
            raise Exception("Could not update")
        resource = self._parse_response(resource)
        return resource
